/*
 * Copyright (C) 2014 digitalSTROM.org, Schlieren, Switzerland
 *
 * digitalSTROM ds485-client library is free software: you can redistribute it
 * and/or modify it under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation, either version 3 of the License,
 * or (at your option) any later version.
 *
 * digitalSTROM ds485-client library is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with digitalSTROM ds485-client library. If not, see
 * <http://www.gnu.org/licenses/>.
 */

#include "ds485-client.h"

#include <assert.h>
#include <errno.h>
#include <pthread.h>
#include <stdarg.h>
#include <string.h>
#include <syslog.h>

#include "base/log.h"
#include "base/socket_client.h"

/* TODO so much code until check level >= logger.threshold */
int dlog(int level, struct _ds485_client_s *cl, const char *fmt, ...)
{
    va_list ap;
    va_start(ap, fmt) ;
    vreport2(level, cl->pretty_name, fmt, ap);
    va_end(ap);
    return 0;
}

void update_pretty_name(ds485_client_t *cl)
{
  int localport;
  if (get_local_port(cl->sockfd, &localport) < 0) {
      localport = -1;
  }
  snprintf(cl->pretty_name, sizeof(cl->pretty_name), "%.15s:%.5s.%05d",
           cl->dName, cl->dPort, localport);
}

int do_container_received(ds485_client_t *cl, ds485_container_t *cnt)
{
    if (cnt->length > DS485_MAX_PAYLOAD_SIZE) {
        dlog(LOG_WARNING, cl, "dropping oversize container: %d\n",
             cnt->length);
        return -EMSGSIZE;
    }

    /* waiting synchronous commands */
    if (waitqueue_waiting(&cl->syncSendShort.wq)) {
        if (dsuid_equal(&cl->syncSendShort.response.sourceId, &cnt->sourceId) &&
            (cl->syncSendShort.response.containerType == cnt->containerType) &&
            (cl->syncSendShort.response.data[0] == cnt->data[0]) &&
            (cl->syncSendShort.response.data[1] == cnt->data[1]))
        {
            cl->syncSendShort.response = *cnt;
            waitqueue_signal(&cl->syncSendShort.wq);
            waitqueue_unlock(&cl->syncSendShort.wq);
            return 0; /* TODO: done really? */
        }
    }
    waitqueue_unlock(&cl->syncSendShort.wq);

    /* Receive Callback */
    struct ds485c_callbacks *callb = &cl->callbacks;
    if (callb->container_pkt_cb) {
        callb->container_pkt_cb(callb->container_pkt_data, cnt);
    }
    return 0;
}

int do_long_container_received(ds485_client_t *cl, ds485n_packet_t *pkt)
{
    if (pkt->length > DS485N_PACKET_MAX_PAYLOAD_SIZE) {
        dlog(LOG_WARNING, cl, "dropping oversize long packet %d\n",
             pkt->length);
        return -EMSGSIZE;
    }

    /* waiting synchronous commands */
    if (waitqueue_waiting(&cl->syncSendLong.wq)) {
        if (dsuid_equal(&cl->syncSendLong.response.sourceId, &pkt->sourceId) &&
            (cl->syncSendLong.response.containerType == pkt->containerType) &&
            (cl->syncSendLong.response.data[0] == pkt->data[0]) &&
            (cl->syncSendLong.response.data[1] == pkt->data[1]))
        {
            cl->syncSendLong.response = *pkt;
            waitqueue_signal(&cl->syncSendLong.wq);
            waitqueue_unlock(&cl->syncSendLong.wq);
            return 0; /* TODO done really? */
        }
    }
    waitqueue_unlock(&cl->syncSendLong.wq);

    /* Receive Callback */
    struct ds485c_callbacks *callb = &cl->callbacks;
    if (callb->netlib_pkt_cb) {
        callb->netlib_pkt_cb(callb->netlib_pkt_data, pkt);
    }
    return 0;
}

int do_link_state_reply_received(ds485_client_t *_cl, bool _connected)
{
    if (waitqueue_waiting(&_cl->busQuery.wq) &&
        _cl->busQuery.type == query_state) {

        _cl->busQuery.length = _connected;
        waitqueue_signal(&_cl->busQuery.wq);
    }
    waitqueue_unlock(&_cl->busQuery.wq);

    /* also emit an event */
    set_link_state(_cl, _connected);

    return 0;
}

static void emit_link_state_event(ds485_client_t *_cl, bool _link)
{
    struct ds485c_callbacks *callb = &_cl->callbacks;
    if (callb->link_cb != NULL) {
        callb->link_cb(callb->link_data, _link);
    }
}

/* triggered by remote daemon message, but also by loss of network connection */
void set_link_state(ds485_client_t *_cl, bool _connected)
{
    bool old = _cl->link;
    _cl->link = _connected;

    if (old != _connected) {
        emit_link_state_event(_cl, _connected);
    }
}

/**
 * do_device_list_received
 * @list - byte string with addresses, not aligned on dsid_t boundaries
 * @len - length of list, multiple of dsid_t
 */
int do_device_list_received(ds485_client_t *cl, uint8_t *octet, int len)
{
    if (waitqueue_waiting(&cl->busQuery.wq) &&
        cl->busQuery.type == query_devices) {

        cl->busQuery.length = len / sizeof(dsuid_t);
        memcpy(cl->busQuery.devices, octet, len);
        waitqueue_signal(&cl->busQuery.wq);
    }
    waitqueue_unlock(&cl->busQuery.wq);
    return 0;
}

/**
 * do_bus_change_received_
 * @cl -- client received packet
 * @dsid -- device that changed, dsid follows memory alignment
 * @flags -- 0 -> joined, 1-> left, see DSM_LEFT
 */
int do_bus_change_received(ds485_client_t *cl, dsuid_t dsuid, int flags)
{
    struct ds485c_callbacks *callb = &cl->callbacks;
    if (callb->bus_change_cb!= NULL) {
        callb->bus_change_cb(callb->bus_change_data, &dsuid, flags);
    }
    return 0;
}

int do_dsid_query_received(ds485_client_t *cl, uint8_t *octet)
{
    if (waitqueue_waiting(&cl->busQuery.wq) &&
        (cl->busQuery.type == query_dsuid)) {

        memcpy(cl->busQuery.devices, octet, sizeof(dsuid_t));
        waitqueue_signal(&cl->busQuery.wq);
    }
    waitqueue_unlock(&cl->busQuery.wq);
    return 0;
}

int do_restart_bus_received(ds485_client_t *cl)
{
    if (waitqueue_waiting(&cl->busQuery.wq) &&
        (cl->busQuery.type == restart_bus)) {
        waitqueue_signal(&cl->busQuery.wq);
    }
    waitqueue_unlock(&cl->busQuery.wq);
    return 0;
}

int do_client_connect_success(ds485_client_t *cl) {
    if (waitqueue_waiting(&cl->openQuery.wq)) {
        waitqueue_signal(&cl->openQuery.wq);
    }
    waitqueue_unlock(&cl->openQuery.wq);
    return 0;
}
